/*	main.cpp
*
*	Game Timer
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <stdio.h>
#include <avr/io.h>
#include "main.h"
#include "Timer.h"
#include "DIP.h"
#include "LED.h"
#include "sound.h"
#include "analog.h"

const uchar cstepLED = 12;		// 12 LEDs around circle

/* 'ms' is a time in milliseconds, based on MCU Timer/Counter [sort of like saying '6:00 p.m.'];
   'dms' is 'delta' millisecond, i.e. a time interval [e.g. 5 minutes] */
const uchar dmsDebounce = 5;	// wait 5 ms after button press or release before look for another change

const ulong dmsBlink = 400;		// approx time for each LED blink when timer is running

/* end-of-time LED flashing */
const ulong dmsFlash = 200;		// time for each flash
const uchar cstepFlash = 10;	// 10 flashes (2 seconds) (minimum) when timer finishes

/*
*	button
*/

#define FBtnUp() (PINB & 0x10)
#define FBtnDown() (!(PINB & 0x10))

/* Flags to indicate Button Status */
const uchar fbsDown     = 0x01;		// set when button is down
const uchar fbsDebounce = 0x02;		// set when during debounce period (button recently pressed or released)

/* BuTtoN value, set by button-checking code for use in states */
enum
{
	btnNil,		// no change to button
	btnPress,	// button just pressed
	btnRelease,	// button just released
};

/*
*	Game timer State
*/

/* 'Init' is for stuff that needs to happen when entering state group;
   'Pending' is when button is down and we're waiting for release */
enum
{
	gsIdleInit,			// entering idle state
	gsIdle,				// in idle state, waiting for button press
	
	gsTimerInit,		// entering timer group
	gsTimerPending,		// waiting for button release
	gsTimerStart,		// start timer!
	gsTimerResume,		// resume timer after pause
	gsTimerNewStep,		// advance to new step (another red LED)
	gsTimerBlink,		// time to blink LED
	gsTimerShowStep,	// set pattern based on step/blink
	gsTimer,			// in timer, watching for time/button event
	
	gsPauseInit,		// entering pause group
	gsPausePending,		// waiting for button release
	gsPause,			// in pause state, waiting for button release

	gsPlayInit,			// entering play group
	gsPlayBlink,		// blink LEDs
	gsPlay,				// playing/flashing, watching for time/button event
};

/*
*	MCU resources
*/

/* timer... 8 us / tick, hi-res up to 572 minutes (max game timer = 5*32 = 160) */
TIMER g_timer;		// TC0 (specified in Config.h)

/* port B has 6 LED lines, button, and buzzer; LED class controls LEDs by setting DDRB and PORTB;
   need to make sure it uses correct DDR for buzzer pin, which is based on whether buzzer is on
   (SOUND class sets / clears appropriate bit in g_ddrbDef) */
uchar g_ddrbDef = 0x00;			// all output at start (buzzer not running)


int main()
{
	DIP dip;
	LED led;
	SOUND sound;	// TC1
	ANALOG analog;

	/* oscillator calibration
	   (osccal value: actual (attempting 60 s) [battery meter: 6 / 12 LEDs])
	   147: 58.46 (def't)
	   130: 62.97
	   141: 59.9 -- good! (59.98 next trial, release code)
	   (very low batt (2 LEDs): 60.14)
	   (fresh alkaline (12 LEDs): 59.93)
	*/
	OSCCAL = 141;
//	OSCCAL = 130;	// for Bob's timer

	/* init digital */
	/* B2 -- buzzer: output, low
	   B4 -- button: input with pull-up
	   rest are LED control lines: tri-state (input, no pull-up) */
	DDRB = 0x04;
	PORTB = 0x10;
	/*  C0-5, D0-3 are DIP: input with pull-up (D4-7 unused; set as low out) */
	DDRC = 0x00;
	PORTC = 0x3f;
	DDRD = 0xf0;
	PORTD = 0x0f;

	if (FBtnDown())
	{
		uchar batt;
		uchar vol;
		
		/* setup mode -- show power level, play tone/tune for volume adjustment / selection */
		led.SetPattern(patternBattLevelLow);
		while (!analog.FBattLevel(&batt))
			led.DoIdle();
		if (batt > cstepLED-1)
			batt = cstepLED-1;
		led.SetPattern(patternBattLevelLow + batt);

		vol = 255;
		while (1)
		{
			if (analog.FVolChange(&vol))
				sound.SetVolume(vol);
			if (sound.FIdle())
				sound.StartSound(dip.Snd());
			led.DoIdle();
			sound.DoIdle();
		}

	}
	else
	{
		uchar gs, grfbs, btn;
		uchar istepLED;
		uchar iblink, cblink;
		uchar fBattLow;
		uchar vol;
		ulong msDebounceStart;
		ulong msTimerStart, msStepStart;
		ulong dmsTimer, dmsEndOfStep, dmsEndOfBlink;

		gs = gsIdleInit;
		grfbs = 0;
		istepLED = 0;
		fBattLow = fFalse;
		vol = 255;
		msTimerStart = msStepStart = msDebounceStart
			= dmsTimer = dmsEndOfStep = dmsEndOfBlink = 0;	// make compiler happy
		iblink = cblink = 0;	// make compiler happy
		while (1)
		{
			/*
			*	handle button -- debounce time, press/release
			*/
			btn = btnNil;
			if (grfbs & fbsDebounce)
			{
				if (g_timer.FElapsed(msDebounceStart, dmsDebounce))
					grfbs &= ~fbsDebounce;
			}
			else
			{
				/* after debounce time; can check for button change */
				if (grfbs & fbsDown)
				{
					if (FBtnUp())	// just released button
					{
						msDebounceStart = g_timer.MsCur();
						grfbs = fbsDebounce;
						btn = btnRelease;
					}
				}
				else
				{
					if (FBtnDown())	// just pressed button
					{
						msDebounceStart = g_timer.MsCur();
						grfbs = fbsDown | fbsDebounce;
						btn = btnPress;
					}
				}
			}

			/*
			*	do action based on current state and
			*	check for state change (based on time or button)
			*/
			switch (gs)
			{
			
			case gsIdleInit:
				led.SetPattern(fBattLow ? patternIdleLowBatt : patternIdle);
				gs = gsIdle;
				// fall through
			case gsIdle:
				if (btn == btnPress)
					gs = gsTimerInit;
				else if (analog.FBattChange(&fBattLow))
					led.SetPattern(fBattLow ? patternIdleLowBatt : patternIdle);
				break;

			case gsTimerInit:
				led.SetPattern(patternStartPending);
				/* set dmsTimer, cblink */
				dmsTimer = dip.MsTimer();
				/* blinks / step = (ms / step) / (ms / blink) [rounded] */
				cblink = ((dmsTimer + cstepLED/2) / cstepLED + dmsBlink/2) / dmsBlink;
				if (cblink % 2 == 0)
					++cblink;	// ensure odd number of blinks; start and end with green
				gs = gsTimerPending;
				// fall through
			case gsTimerPending:
				if (btn == btnRelease)
					gs = gsTimerStart;
				break;
			case gsTimerStart:
				msTimerStart = g_timer.MsCur();
				istepLED = 0;
				// fall through
			case gsTimerNewStep:	// time for a new step
				/* record start time for step (used when blinking), increment step,
				   calc time for end of step */
				msStepStart = g_timer.MsCur();
				/* note that when we're determining the time for an interval (next LED step
				   or blink), we don't want to use a simple ms/count calculation because
				   truncation can cause errors to accumulate; we need to look at where we
				   should be at the end of each interval as a proportion of the whole */
				/* calculate time until the end of this step (from beginning of timer) */
				dmsEndOfStep = (dmsTimer * ++istepLED + cstepLED/2) / cstepLED;
				iblink = 0;
				// fall through
			case gsTimerBlink:		// time to blink
				/* calculate time until the end of this blink (from beginning of step) */
				dmsEndOfBlink = ((dmsEndOfStep - (msStepStart - msTimerStart)) * ++iblink
					+ cblink/2) / cblink;
				// fall through
			case gsTimerShowStep:
				led.SetPattern(patternStart + istepLED - ((iblink % 2 == 0) ? 0 : 1));
				gs = gsTimer;
				// fall through
			case gsTimer:
				if (btn == btnPress)
				{
					gs = (dip.FReset() ? gsIdleInit : gsPauseInit);
				}
				else
				{
					if (g_timer.FElapsed(msTimerStart, dmsEndOfStep))
						gs = istepLED >= cstepLED ? gsPlayInit : gsTimerNewStep;
					else if (g_timer.FElapsed(msStepStart, dmsEndOfBlink))
						gs = gsTimerBlink;
					if (istepLED >= cstepLED)
					{
						/* working on the last step (or just finished); start checking volume */
						if (analog.FVolChange(&vol))
							sound.SetVolume(vol);
					}
				}
				break;
			case gsTimerResume:
				/* adjust start time to account for time elapsed before pause */
				msTimerStart = g_timer.MsCur() - msTimerStart;
				msStepStart = g_timer.MsCur() - msStepStart;
				/* (istepLED, iblink, cblink, dmsTimer, and dmsEndOfTimer/Step still set from before pause) */
				Assert(istepLED > 0);
				Assert(iblink > 0);
				gs = gsTimerShowStep;
				break;

			case gsPauseInit:
				{
					uchar d;
					ulong msT;
						
					/* borrow 'start time' vars to store elapsed time so far */
					msT = g_timer.MsCur();
					msTimerStart = msT - msTimerStart;
					msStepStart = msT - msStepStart;

					/* show indication of stop/pause and current elapsed time */
					/* istepLED=1 is all green, 0 red (flashing 1 red); =2 is 1 red, etc;
					   batt level starts with 1 (red) LED */
					d = istepLED >= 2 ? istepLED - 2 : 0;
					led.SetPattern(patternBattLevelLow + d);
					gs = gsPausePending;
				}
				// fall through
			case gsPausePending:
				if (btn == btnRelease)
					gs = gsPause;
				break;
			case gsPause:
				if (btn == btnRelease)
					gs = gsTimerResume;
				break;

			case gsPlayInit:
				sound.StartSound(dip.Snd());
				cblink = cstepFlash;	// number of times to flash
				// fall through
			case gsPlayBlink:
				msTimerStart = g_timer.MsCur();		// timer is now just measuring blink time
				led.SetPattern(patternFlash1 + cblink % 2);
				gs = gsPlay;
				// fall through
			case gsPlay:
				/* if button is pressed while playing, we stop the tune and go back to idle */
				if (btn == btnPress)
				{
					sound.Stop();
					gs = gsIdleInit;
				}
				else
				{
					if (g_timer.FElapsed(msTimerStart, dmsFlash))
					{
						/* time to flash! */
						if (--cblink == 0)
						{
							/* done flashing; see if still playing tune */
							if (sound.FIdle())
							{
								gs = gsIdleInit;
								break;
							}
							cblink = 2;	// flash some more
						}
						gs = gsPlayBlink;
					}
					if (analog.FVolChange(&vol))
						sound.SetVolume(vol);
				}
				break;
				
			default:
				Assert(fFalse);
			}

			/*
			*	do idle processing for LEDs (Charlieplexing) and sound
			*/
			led.DoIdle();
			sound.DoIdle();
		}
	}

	return 0;
}
