/*	sound.cpp
*
*	Game Timer buzzer, uses Timer/Counter 1
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "main.h"
#include "sound.h"

/* buzzer connected to PB2 / OC1B */
const uchar fpinBuzzer = 0x04;

/*
WGM 1110: fast PWM, top=ICR1
COM1B 10: clear on match, set at BOTTOM
CS 101: 8 MHz / 1024
CS 100: 8 MHz / 256
CS 011: 8 MHz / 64
CS 010: 8 MHz / 8
CS 001: 8 MHz / 1
*/

const uchar tccr1aBase = 0x02;	// 00 00 .. 10 (no output [yet], WGM low 2 bits)
const uchar tccr1bBase = 0x18;	// 00. 11 000 (WGM high 2 bits, cs disabled [for now])
const uchar tccr1aRun = tccr1aBase | 0x20;	// .. 10 .. .. (COM1B 10: clear on match, set on bottom)
const uchar tccr1bRun = tccr1bBase | 0x02;	// .... . 010 (CS 010: clk/8)

/*
fixed point numbers...
numerator stored in short, denominator = 2^16 ==> fraction range [0 - 1)
add & subtract are easy
multiplication...
    x      y     x * y   x*y / 2^16
   ---- * ---- = ----- = ----------              so, just mult as long, then take high short
   2^16   2^16   2^32       2^16
division...
    x      y     x   x/y * 2^16   x*2^16 / y
   ---- / ---- = - = ---------- = ----------     so, shift x to high short, then divide by y
   2^16   2^16   y      2^16         2^16
*/
const ulong denomFixedPt = 0x10000;		// denominator for scale value

/* for volume calc */
const uchar volMax = 160;	// max expected from voltage divider (50k pot to GND, 30k R to Vcc)
const ushort kvolume = (ushort)(0.965 * denomFixedPt + 0.5);	// (+0.5 to round)

/* sndTune */
const NOTE s_anote1812[] PROGMEM =
{
	{2550, 200}, {0, 20},
	{1910, 200}, {0, 20},
	{1702, 200}, {0, 20},
	{1516, 200}, {0, 20},
	{1702, 200}, {0, 20},
	{1910, 200}, {0, 20},
	{1702, 200}, {0, 20},
	{1516, 400}, {0, 20},
	{1910, 400}, {0, 20},
	{1910, 800}, {0, 20},
};
const uchar cnote1812 = DIM(s_anote1812);

/* sndBeeps */
const uchar cbeep = 15;
const ushort topBeep = 955;
const ushort dmsBeep = 150;
const ushort dmsBeepPause = 10;

/* sndSiren */
const uchar csiren = 6;
const ushort dmsSiren = 14;
const ushort topMostSiren = 955;	// lowest note
const ushort topMinSiren = 477;		// highest note
const ushort ksirenRise = (ushort)(0.971531941 * denomFixedPt + 0.5);	// (+0.5 to round)
const ulong ksirenFall = (ulong)(1.029302237 * denomFixedPt + 0.5);	// (+0.5 to round)


SOUND::SOUND() : m_vol((ushort)(denomFixedPt/2)), m_snd(sndNil), m_pnote(NULL), m_cnote(0)
{
	TCCR1A = tccr1aBase;	
	TCCR1B = tccr1bBase;	
}

/* this function calculates an appropriate scale factor based on the given volume; when sound
   is played, the scale factor is used to calculate the width of the pulse, which can range
   from 1/2 (50% pulse width) for vol=volMax (loudest) down to a small fraction for vol=0 */
void SOUND::SetVolume(uchar vol)
{
	ushort base;
	uchar exp;
	ulong result;

	if (vol > volMax)
		vol = volMax;

	/* since perceived volume is based on a logarithmic scale, need to use an exponential to
	   scale the pulse width; in this case "kvolume" (constant set above) raised to a power,
	   with lower volumes using larger powers (more reduction in sound), so:
	   
	             (volMax - vol)
	      kvolume
	*/

	/* Here is an efficient way to calculate the (intergral) power of a number.

	   Consider the equation:

	       y     1     2     4     8       128
	      x  = ax  * bx  * cx  * dx  ... hx

	   Each of a, b, ... h will be either 0 or 1 as required so that the
	   exponents of the '1' terms sum to y. The exponents are powers of 2.

	   For example:
	       11   1    2    8
	      x  = x  * x  * x
	   Notice that the coefficient of the x^4 term is 0 (as with all exponents > 8).

	   The algorithm...
	      The value of each x^n term is calculated as we go by squaring the previous term
	         (e.g. x^16 = x^8 * x^8).
	      Look at the bits of the exponent to determine which terms to include.
	         (The bits correspond to the coefficients a, b, c, etc.)
	*/

	base = kvolume;
	exp = volMax - vol;
	result = denomFixedPt;	// result = 1 in our fixed point system
	while (exp > 0)
	{
		if (exp & 1)
			result = (result * base) >> 16;			// fixed point: result * base

		base = (ushort)(((ulong)base * base) >> 16);	// fixed point: base * base
		exp >>= 1;
	}

	/* 'result' now has an appropriate fraction from 1 (loud) down to a small value (quiet);
	   this is the amount by which we need to adjust the 50% pulse width (loudest) for
	   the specified volume; smaller numbers for result will result in shorter pulses and thus
	   quieter volume */
	m_vol = (ushort)(result / 2);	// result x 50%
}

void SOUND::StartSound(uchar snd)
{
	if (m_snd != sndNil)
		StopOutput();

	switch (m_snd = snd)
	{
	case sndTune:
		m_pnote = s_anote1812;
		m_cnote = cnote1812;
		break;
	case sndBeeps:
		m_cbeep = cbeep * 2;	// tone + pause for each 'beep'
		break;
	case sndSiren:
		m_csiren = csiren;
		m_top = topMostSiren;	// largest value for Timer/Counter top = lowest freq = lowest note
		m_dms = dmsSiren;
		break;
	default:
		m_snd = sndNil;
		return;
	}

	StartNote();
}

void SOUND::DoIdle()
{
	uchar fMore;
	
	if (m_snd != sndNil)
	{
		if (g_timer.FElapsed(m_msStart, m_dms))
		{
			StopOutput();
			
			/* advance to next part of sound */
			fMore = fFalse;
			switch (m_snd)
			{
			case sndTune:
				if (--m_cnote > 0)
				{
					++m_pnote;
					fMore = fTrue;
				}
				break;
			case sndBeeps:
				fMore = --m_cbeep > 0;
				break;
			case sndSiren:
				if (m_csiren & 0x01)	// falling sound = increasing 'top'
				{
					m_top = (ushort)(((ulong)m_top * ksirenFall) >> 16);
					if (m_top <= topMostSiren)
						fMore = fTrue;
					else if (--m_csiren > 0)
					{
						m_top = topMostSiren;
						fMore = fTrue;
					}
				}
				else					// rising sound = decreasing 'top'
				{
					m_top = (ushort)(((ulong)m_top * ksirenRise) >> 16);
					if (m_top >= topMinSiren)
						fMore = fTrue;
					else if (--m_csiren > 0)
					{
						m_top = topMinSiren;
						fMore = fTrue;
					}
				}
				break;
			}

			if (fMore)
				StartNote();
			else
				m_snd = sndNil;
		}
	}
}

void SOUND::Stop()
{
	StopOutput();
	m_snd = sndNil;
}

void SOUND::StartNote()
{
	NOTE note;

	switch (m_snd)
	{
	case sndTune:
		memcpy_P(&note, m_pnote, sizeof(NOTE));
		m_dms = note.dms;
		ICR1 = note.top;
		break;
	case sndBeeps:
		/* beep on even, pause on odd */
		if (m_cbeep & 0x01)
		{
			ICR1 = 0;
			m_dms = dmsBeepPause;
		}
		else
		{
			ICR1 = topBeep;
			m_dms = dmsBeep;
		}
		break;
	case sndSiren:
		ICR1 = m_top;
		/* m_dms set in StartSound; unchanged */
		break;
	}
	
	if (ICR1 != 0)
	{
		TCNT1 = 0;
		OCR1B = (ushort)((ulong)m_vol * ICR1 / denomFixedPt);
		StartOutput();
	}
	/* else pause; leave timer output disconnected */

	m_msStart = g_timer.MsCur();
}

void SOUND::StartOutput()
{
	g_ddrbDef |= fpinBuzzer;
	DDRB |= fpinBuzzer;
	TCCR1A = tccr1aRun;
	TCCR1B = tccr1bRun;
}

void SOUND::StopOutput()
{
	TCCR1A = tccr1aBase;
	TCCR1B = tccr1bBase;
	DDRB &= ~fpinBuzzer;
	g_ddrbDef &= ~fpinBuzzer;
}
